/** @file cflash.cpp
 *  @author Mic
 *	@date 2007
 *  @brief CompactFlash/FAT emulation implementation for Dualis.
 *   
 *	Logical memory layout:\n
 *  <code>
 *	----------------------\n
 *	|        MBR         |\n
 *	----------------------\n
 *	|        FAT         |\n
 *	----------------------\n
 *	|    Root entries    |\n
 *	----------------------\n
 *	|   Subdirectories   |\n
 *	----------------------\n
 *	|     File data      |\n
 *	----------------------\n
 * </code>
 */


#include "cflash.h"
#include <stdio.h>


#define BYTESPERSEC 512
#define SECPERFAT	128
#define SECPERCLUS	16
#define MAXFILES	32768
#define SECRESV		2
#define NUMSECTORS	0x80000
#define NUMCLUSTERS	(NUMSECTORS/SECPERCLUS)
#define BYTESPERCLUS (BYTESPERSEC*SECPERCLUS)
#define DIRENTSPERCLUS ((BYTESPERCLUS)/32)


unsigned short				cf_reg_data,cf_reg_err,cf_reg_sec,cf_reg_lba1,cf_reg_lba2,
							cf_reg_lba3,cf_reg_lba4,cf_reg_cmd,cf_reg_sts;
unsigned int				CF_REG_DATA,CF_REG_ERR,CF_REG_SEC,CF_REG_LBA1,CF_REG_LBA2,
							CF_REG_LBA3,CF_REG_LBA4,CF_REG_CMD,CF_REG_STS;

BOOT_RECORD					MBR;
DIR_ENT						*files,*dirEntries,**dirEntryPtr;
FILE_INFO					*fileLink,*dirEntryLink;
UINT						currLBA;
UINT						filesysFAT,filesysRootDir,filesysData;
u16							FAT16[SECPERFAT*256];
u16							numExtraEntries[SECPERFAT*256];
DIR_ENT						*extraDirEntries[SECPERFAT*256];
int							numFiles,fileLevel,maxLevel,dirNum,numRootFiles;
int							*dirEntriesInCluster,clusterNum,
							firstDirEntCluster,lastDirEntCluster,
							lastFileDataCluster;
HANDLE						houtF;
char						*sRomPath;
int							activeDirEnt=-1;
UINT						bufferStart;
UINT						fileStartLBA,fileEndLBA;
unsigned short				freadBuffer[256];
DWORD						dwBytesRead;
HANDLE						hFile;
char						fpath[MAX_PATH+1];
bool						cflashDeviceEnabled = false;



/** 
 *  @brief Calculates the FAT format long filename checksum for last file added
 *  @see add_file
 *
 *  @return The LFN checksum
 */
static int lfn_checksum() {
	int i;
	unsigned char chk;

	chk = 0;
	for (i=0; i < 11; i++) {
		chk = ((chk & 1) ? 0x80 : 0) + (chk >> 1) + (i < 8 ? files[numFiles].name[i] : files[numFiles].ext[i - 8]);
	}
	return chk;
}


const int lfnPos[13] = {1,3,5,7,9,14,16,18,20,22,24,28,30};


/** 
 *  @brief Adds a DIR_ENT for the files.
 *
 *  @return
 */
void add_file(char *fname, WIN32_FIND_DATA *FindFileData) {
	int i,j,k,n;
	unsigned char chk;
	char *p;

	if (!fname || !FindFileData)
		return;

	if (numFiles < MAXFILES-1) {
		if (strcmp(fname, "..") != 0) {
			for (i = strlen(fname)-1; i>=0 && (fname[i] != '.'); i--) {}
			if ((i == 0) && (strcmp(fname,".") == 0)) i = 1;	// Handle the case where the filename is a single dot (i.e. the current dir) 
			if (i < 0) i = strlen(fname);						// Handle the case where the file has no extension 	
			for (j=0; j<i; j++)
				files[numFiles].name[j] = fname[j];			// Copy the name
			for (; j<8; j++)
				files[numFiles].name[j] = 0x20;				// Pad with spaces as needed
			for (j=0; j<3; j++) {
				if ((j+i+1)>=strlen(fname)) break;
				files[numFiles].ext[j] = fname[j+i+1];		// Copy the extension
			}
			for (; j<3; j++)
				files[numFiles].ext[j] = 0x20;				// Pad with spaces as needed

			fileLink[maxLevel * ((fileLevel>0)?1:0)].filesInDir += 1;	// Increase the counter for number of files in this directory

			// See if LFN entries need to be added
			if (strlen(FindFileData->cAlternateFileName) > 0) {
				chk = lfn_checksum();
				// Calculate the number of dirents needed for the filename
				k = (strlen(FindFileData->cFileName)/13) + (((strlen(FindFileData->cFileName)%13)!=0)?1:0);
				numFiles += k;

				// Update the file counter
				fileLink[maxLevel * ((fileLevel>0)?1:0)].filesInDir += k;

#ifdef DEBUG
				wsprintf(buffer,"Adding %d LFN entries for file %s\n",k,fname);
				WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
#endif				
				n = 0;
				j = 13;
				for (i=0; i<strlen(FindFileData->cFileName); i++) {
					if (j == 13) {
						n++;
						p = (char*)&files[numFiles-n].name[0];
						fileLink[numFiles-n].parent = maxLevel * ((fileLevel > 0) ? 1 : 0);	
						p[0] = n;
						((DIR_ENT*)p)->attrib = ATTRIB_LFN;
						p[0xD] = chk;
						j = 0;
					}
					*(p + lfnPos[j]) = FindFileData->cFileName[i];
					*(p + lfnPos[j]+1) = 0;
					j++;
				}
				for (; j<13; j++) {
					*(p + lfnPos[j]) = FindFileData->cFileName[i];
					*(p + lfnPos[j]+1) = 0;
				}
				p[0] |= 0x40;	// END
				for (i=strlen(fname)-1; i>=0; i--)
					if (fname[i]=='.') break;
				if ((i==0)&&(strcmp(fname,".")==0)) i = 1;
				if (i<0) i = strlen(fname);
				for (j=0; j<i; j++)
					files[numFiles].name[j] = fname[j];
				for (; j<8; j++)
					files[numFiles].name[j] = 0x20;
				for (j=0; j<3; j++) {
					if ((j+i+1)>=strlen(fname)) break;
					files[numFiles].ext[j] = fname[j+i+1];
				}
				for (; j<3; j++)
					files[numFiles].ext[j] = 0x20;
			}

			files[numFiles].fileSize = FindFileData->nFileSizeLow;

			if (FindFileData->dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
				if (strcmp(fname,".")==0)
					fileLink[numFiles].level = maxLevel;
				else
					fileLink[numFiles].level = maxLevel+1;
				files[numFiles].attrib = ATTRIB_DIR;
			} else {
				files[numFiles].attrib = 0;
			}

			fileLink[numFiles].parent = maxLevel * ((fileLevel>0)?1:0);	

			numFiles++;

		} else if (fileLevel > 0) {
			fileLink[maxLevel * ((fileLevel > 0) ? 1 : 0)].filesInDir += 1;
			strncpy_s((char*)&files[numFiles].name[0], 10, "..      ", 8);
			strncpy_s((char*)&files[numFiles].ext[0], 4, "   ", 3);
			fileLink[numFiles].parent = maxLevel * ((fileLevel > 0) ? 1 : 0);	
			files[numFiles].attrib = 0x10;
			numFiles++;
		}
	}
}


/** 
 *  @brief Lists all files and subdirectories recursively
 *
 *	@param pszPath Pointer to a string containing the name of the directory to list
 *
 *  @return
 */
void list_files(const char *pszPath) {
	WIN32_FIND_DATA	FindFileData;
	HANDLE			hFind = INVALID_HANDLE_VALUE;
	char			DirSpec[MAX_PATH + 1],SubDir[MAX_PATH+1],fnameu[256];
	DWORD			dwError;
	char			*fname;
	int				i,j;

	if (!pszPath)
		return;

	fileLevel++;
	maxLevel++;
	
	strncpy_s(DirSpec, MAX_PATH, pszPath, strlen(pszPath)+1);
	strncat_s(DirSpec, MAX_PATH, "\\*", 3);

	hFind = FindFirstFile(DirSpec, &FindFileData);

	if (hFind == INVALID_HANDLE_VALUE) {
		wsprintf(buffer,
			     "Invalid file handle. Error is %u\n",
				 GetLastError());
		WriteConsole(hout,
			         buffer,
					 lstrlen(buffer),
					 (unsigned long*)&dummy,
					 NULL);
		return;
	} else {
		fname = (strlen(FindFileData.cAlternateFileName) > 0) ? FindFileData.cAlternateFileName : FindFileData.cFileName;
		strcpy_s(fnameu, 256, fname);
		_strupr(fnameu);
		add_file(fnameu, &FindFileData);

		// Repeat as long as there are more files to list
		while (FindNextFile(hFind, &FindFileData) != 0) {
			// get the filename and convert it to upper case
			fname = (strlen(FindFileData.cAlternateFileName) > 0) ? FindFileData.cAlternateFileName : FindFileData.cFileName;
			strcpy_s(fnameu, 256, fname);
			_strupr(fnameu);
			
			// Create one or more dirents for the file
			add_file(fnameu, &FindFileData);

			// Quit if we've used up all the dirents
			if (numFiles == MAXFILES-1) break;

			// Check if this was a directory, and if so scan it as well
			if ((FindFileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)&&
			   (fname[0] != '.')) {
				wsprintf(SubDir,"%s\\%s",pszPath,fname);
				list_files(SubDir);
			}
		}
    
		dwError = GetLastError();
		FindClose(hFind);
		if (dwError != ERROR_NO_MORE_FILES) {
			wsprintf(buffer,"FindNextFile error. Error is %u\n", dwError);
			WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
			return;
		}
	}
	if (numFiles < MAXFILES) {
		fileLink[numFiles].parent = maxLevel * ((fileLevel > 0) ? 1 : 0);
		files[numFiles++].name[0] = 0;
	}
	fileLevel--;
}






/** 
 *  @brief Sets up the MBR, FAT and DIR_ENTs
 *
 *	@return true on success, otherwise false
 */
bool cflash_build_fat() {
	int i,j,k,l,
		clust,baseCluster,numClusters,
		clusterNum2,rootCluster;

	numFiles  = 0;
	fileLevel = -1;
	maxLevel  = -1;

	sRomPath  = (char*)zombie(ASK,ZOMBIE_ROM_PATH, 0L, 0L);
	
	if ((files = (DIR_ENT*)LocalAlloc(LPTR,MAXFILES * sizeof(DIR_ENT))) == NULL)
		return false;
	if ((fileLink  = (FILE_INFO*)LocalAlloc(LPTR,MAXFILES * sizeof(FILE_INFO))) == NULL) {
		LocalFree(files);
		return false;
	}

	for (i = 0; i < MAXFILES; i++) {
		files[i].attrib = 0;
		files[i].name[0] = FILE_FREE;
		files[i].fileSize = 0;

		fileLink[i].filesInDir = 0;

		extraDirEntries[i] = NULL;
		numExtraEntries[i] = 0;
	}

	list_files(sRomPath);

	wsprintf(buffer,
		    "Found %d files\n",
			numFiles);
	WriteConsole(hout,
		         buffer,
				 lstrlen(buffer),
				 (unsigned long*)&dummy,
				 NULL);

	k            = 0;
	clusterNum   = rootCluster = (SECRESV + SECPERFAT)/SECPERCLUS;
	numClusters  = 0;
	clust        = 0;
	numRootFiles = 0;

	// Allocate memory to hold information about the files 
	dirEntries          = (DIR_ENT*)LocalAlloc(LPTR,numFiles*sizeof(DIR_ENT));
	dirEntryLink        = (FILE_INFO*)LocalAlloc(LPTR,numFiles*sizeof(FILE_INFO));
	dirEntriesInCluster = (int*)LocalAlloc(LPTR,NUMCLUSTERS*sizeof(int));
	dirEntryPtr         = (DIR_ENT**)LocalAlloc(LPTR,NUMCLUSTERS*sizeof(DIR_ENT*));
	if ((dirEntries==NULL) || (dirEntriesInCluster==NULL) || (dirEntryPtr==NULL))
		return false;

	for (i=0; i<NUMCLUSTERS; i++) {
		dirEntriesInCluster[i] = 0;
		dirEntryPtr[i] = NULL;
	}

	/*for (i=0; i<8; i++) {
		wsprintf(buffer,"%d dirents in dir %d\n",fileLink[i].filesInDir,i);
		WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
	}*/

	// Change the hierarchical layout to a flat one 
	for (i=0; i<=maxLevel; i++) {
		clusterNum2 = clusterNum;
		for (j=0; j<numFiles; j++) {
			if (fileLink[j].parent == i) {
				if (dirEntryPtr[clusterNum] == NULL)
					dirEntryPtr[clusterNum] = &dirEntries[k];
				dirEntryLink[k] = fileLink[j];
				dirEntries[k++] = files[j];
				if ((files[j].attrib & ATTRIB_LFN) == 0) {
					if (files[j].attrib & ATTRIB_DIR) {
						if (strncmp((char*)&files[j].name[0], ".       ", 8) == 0) {
							dirEntries[k-1].startCluster = dirEntryLink[k-1].level; 

						} else if (strncmp((char*)&files[j].name[0], "..      ", 8) == 0) {
							dirEntries[k-1].startCluster = dirEntryLink[k-1].parent; 
						
						} else {
							clust++;
							dirEntries[k-1].startCluster = clust;
							l = (fileLink[fileLink[j].level].filesInDir)>>8;
							clust += l;
							numClusters += l;
						}

					} else {
						dirEntries[k-1].startCluster = clusterNum;
					}
				}

				if (i==0)
					numRootFiles++;
				dirEntriesInCluster[clusterNum]++;
				if (dirEntriesInCluster[clusterNum] == 256) {
					clusterNum++;
				}
			}
		}
		clusterNum = clusterNum2 + ((fileLink[i].filesInDir)>>8) + 1;
		numClusters++;
	}

	// Free the file indexing buffer
	LocalFree((HLOCAL)files);
	LocalFree((HLOCAL)fileLink);

	// Set up the MBR
	MBR.bytesPerSector = BYTESPERSEC;
	MBR.numFATs = 1;
	strcpy_s((char*)&MBR.OEMName[0], 8, "DUALIS");
	strcpy_s((char*)&MBR.fat16.fileSysType[0], 8, "FAT16  ");
	MBR.reservedSectors = SECRESV;
	MBR.numSectors = 524288;
	MBR.numSectorsSmall = 0;
	MBR.sectorsPerCluster = SECPERCLUS;
	MBR.sectorsPerFAT = SECPERFAT;
	MBR.rootEntries = 512;
	MBR.fat16.signature = 0xAA55;
	MBR.mediaDesc = 1;

	filesysFAT = 0 + MBR.reservedSectors;
	filesysRootDir = filesysFAT + (MBR.numFATs * MBR.sectorsPerFAT);
	filesysData = filesysRootDir + ((MBR.rootEntries * sizeof(DIR_ENT)) / BYTESPERSEC);

	// Set up the cluster values for all subdirectories
	clust = filesysData / SECPERCLUS;
	firstDirEntCluster = clust;
	for (i=1; i<numFiles; i++) {
		if (((dirEntries[i].attrib & ATTRIB_DIR)!=0) &&
			((dirEntries[i].attrib & ATTRIB_LFN)==0)) {
			if (dirEntries[i].startCluster > rootCluster)
				dirEntries[i].startCluster += clust-rootCluster;
		}
	}
	lastDirEntCluster = clust+numClusters-1;

	// Set up the cluster values for all files
	clust += numClusters; //clusterNum;
	for (i=0; i<numFiles; i++) {
		if (((dirEntries[i].attrib & ATTRIB_DIR)==0) &&
			((dirEntries[i].attrib & ATTRIB_LFN)==0)) {
			dirEntries[i].startCluster = clust;
			clust += (dirEntries[i].fileSize/(BYTESPERSEC*SECPERCLUS));
			clust++;
			//if (i<20) {
			//	wsprintf(buffer,"Starting cluster of file %d is %d\n",i,dirEntries[i].startCluster);
			//	WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
			//}
		}
	}
	lastFileDataCluster = clust-1;

	// Set up the FAT16
	memset(FAT16,0,SECPERFAT*256*sizeof(u16));
	FAT16[0] = 0xFF01;
	FAT16[1] = 0xFFFF;
	for (i=2; i<=lastDirEntCluster; i++)
		FAT16[i] = 0xFFFF;
	k = 2;
	for (i=0; i<numFiles; i++) {
		if (((dirEntries[i].attrib & ATTRIB_LFN)==0) &&
			(dirEntries[i].name[0] != FILE_FREE)) {
			j = 0;
			l = dirEntries[i].fileSize - (BYTESPERSEC*SECPERCLUS);
			while (l > 0) {
				if (dirEntries[i].startCluster+j < MAXFILES)
					FAT16[dirEntries[i].startCluster+j] = dirEntries[i].startCluster+j+1;
				j++;
				l -= (BYTESPERSEC*16);
			} 
			if ((dirEntries[i].attrib & ATTRIB_DIR)==0) {
				if (dirEntries[i].startCluster+j < MAXFILES)
					FAT16[dirEntries[i].startCluster+j] = 0xFFFF;
			}
			k = dirEntries[i].startCluster+j;
		}
	}

	for (i=(filesysData/SECPERCLUS); i<NUMCLUSTERS; i++) {
		if (dirEntriesInCluster[i]==256)
			FAT16[i] = i+1;
	}

	/*for (i=0; i<16; i++) {
		wsprintf(buffer,"FAT[%d] = %04X\n",i,FAT16[i]);
		WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
	}*/

	/*for (i=0; i<numFiles; i++) {
		wsprintf(buffer,"[%s][%s] %d, cluster=%d\n\n",dirEntries[i].name,dirEntries[i].ext,dirEntries[i].cDate,dirEntries[i].startCluster);
		WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
	}*/

	return true;
}



/** 
 *  @brief Initializes CFlash emulation
 *
 *	@return true on success, otherwise false
 */
bool cflash_init() {
	cflashDeviceEnabled = false;
	currLBA = 0;
	if (activeDirEnt != -1)
		CloseHandle(hFile);
	activeDirEnt = -1;
	fileStartLBA = fileEndLBA = 0xFFFFFFFF;
	if (!cflash_build_fat())
		return false;
	cf_reg_sts = 0x58;	// READY

	cflashDeviceEnabled = true;
	return true;
}


/** 
 *  @brief Converts a space-padded 8.3 filename into an asciiz string
 *
 *  @param      dirent Directory entry number, used to resolve LFNs.
 *  @param[out] pszOut Pointer to string where the result is to be stored.
 *  @param      pEntry Pointer to directory entry, used to resolve LFNs.
 *
 *  @return 
 */
static void fatstring_to_asciiz(int dirent, char *pszOut, DIR_ENT *pEntry) {
	int i,j;
	DIR_ENT *pd;

	if (pEntry == NULL)
		pd = &dirEntries[dirent];
	else
		pd = pEntry;

	for (i = 0; i < 8; i++) {
		if (pd->name[i] == ' ') break;
		pszOut[i] = pd->name[i];
	}
	if ((pd->attrib & 0x10) == 0) {
		pszOut[i++] = '.';
		for (j=0; j<3; j++) {
			if (pd->ext[j] == ' ') break;
			pszOut[i++] = pd->ext[j];
		}
	}
	pszOut[i] = '\0';	// Add null terminator
}



/** 
 *  @brief Resolves the path of a dirent by working backwards through the directory entries.
 *
 *  @param dirent Directory entry number
 *
 *	@return
 */
static void resolve_path(int dirent) {
	int i;
	char dirname[128];

	// Stop when the parent is the root directory
	while (dirEntryLink[dirent].parent > 0) {
		for (i=0; i<dirent; i++) {
			// Check if we've found the parent directory of our dirent
			if ((dirEntryLink[dirent].parent == dirEntryLink[i].level) &&
				((dirEntries[i].attrib&ATTRIB_DIR)!=0)) {
				// Convert the filename to plain asciiz, and concatenate it to what we've
				// already resolved.
				fatstring_to_asciiz(i, dirname, NULL);
				strcat_s(fpath, MAX_PATH, dirname);
				strcat_s(fpath, MAX_PATH, "\\");
				dirent = i;
				break;
			}
		}
	}
}


/** 
 *  @brief Reads from a file using a 512 byte buffer
 *
 *  @param dirent  Directory entry number.
 *  @param cluster Cluster to read from.
 *  @param offset  Offset within the cluster.
 *
 *  @return The 16-bits pointed out by cluster+offset
 */
static unsigned short fread_buffered(int dirent,UINT cluster,UINT offset) {
	char fname[32];
	int i,j;

	offset += cluster * BYTESPERSEC * SECPERCLUS;

	// Check if the file is already open
	if (dirent == activeDirEnt) {
		if ((offset < bufferStart) || (offset >= bufferStart + 512)) {
			SetFilePointer(hFile,
				           offset,
						   NULL,
						   FILE_BEGIN);
			ReadFile(hFile,
				     &freadBuffer,
					 512,
					 &dwBytesRead,
					 NULL);
			bufferStart = offset;
		}

		return freadBuffer[(offset-bufferStart)>>1];
	}
	if (activeDirEnt != -1)
		CloseHandle(hFile);

	strcpy(fpath,sRomPath);
	strcat(fpath,"\\");
	
	resolve_path(dirent);

	fatstring_to_asciiz(dirent,fname,NULL);
	strcat(fpath,fname);

	wsprintf(buffer, "Opening [%s] for reading\n", fpath);
	WriteConsole(hout, buffer, lstrlen(buffer), (unsigned long*)&dummy, NULL);

	hFile = CreateFile(fpath,
		               GENERIC_READ,
					   0,
					   NULL,
					   OPEN_EXISTING,
	                   FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN,
		               NULL);
	if (hFile == INVALID_HANDLE_VALUE)
		return 0;
	//SetFilePointer(hFile,offset,NULL,FILE_BEGIN);
	ReadFile(hFile,
		     &freadBuffer,
			 512,
			 &dwBytesRead,
			 NULL);

	bufferStart = offset;
	activeDirEnt = dirent;
	fileStartLBA = (dirEntries[dirent].startCluster * BYTESPERSEC * SECPERCLUS);
	fileEndLBA = fileStartLBA + dirEntries[dirent].fileSize;

	//wsprintf(buffer,"offset: %d, dirent: %d, startlba: %d, endlba: %d\n",offset,dirent,fileStartLBA,fileEndLBA);
	//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);

	return freadBuffer[(offset-bufferStart)>>1];
}


FILE *gpfWrite = NULL;

/** 
 *  @brief Writes to a file
 *
 *  @return
 */
static void fwrite_buffered(int dirent, DIR_ENT *d, UINT cluster, UINT offset, int chain, int data) {
	static char fname[MAX_PATH];
	int i,j;

	if (!d)
		return;

	offset += cluster * BYTESPERSEC * SECPERCLUS;

	if (dirent == activeDirEnt) {
		if (offset >= 0x200) {
		//wsprintf(buffer,"Writing %X to file pos %X\n",data,offset);
		//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
		}

		fseek(gpfWrite, offset, SEEK_SET);
		fwrite(&data, 2, 1, gpfWrite);
		return;
	}

	/*if (activeDirEnt != -1)
		CloseHandle(hFile);*/

	strcpy(fpath, sRomPath);
	strcat(fpath, "\\");
	
	//resolve_path(dirent);

	fatstring_to_asciiz(dirent, fname, d);
	strcat(fpath, fname);

	wsprintf(buffer, "Opening [%s] for writing\n", fpath);
	WriteConsole(hout,
		         buffer,
				 lstrlen(buffer),
				 (unsigned long*)&dummy,
				 NULL);

	//wsprintf(buffer,"Writing %X to file pos %X\n",data,offset);
	//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);

	fopen_s(&gpfWrite, fpath, "wb");
	fseek(gpfWrite, offset, SEEK_SET);
	fwrite(&data, 2, 1,gpfWrite);

	/*hFile = CreateFile(fpath, GENERIC_READ, 0, NULL, OPEN_EXISTING,
	                   FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN,
		               NULL);
	if (hFile == INVALID_HANDLE_VALUE)
		return 0;
	ReadFile(hFile,&freadBuffer,512,&dwBytesRead,NULL);*/

	bufferStart = offset;
	activeDirEnt = dirent;
	fileStartLBA = (dirEntries[dirent].startCluster * BYTESPERSEC * SECPERCLUS);
	fileEndLBA = fileStartLBA + dirEntries[dirent].fileSize;
}



/**
 *  @brief Reads from one of the CF device registers.
 *
 *  @param address Address to read from. Addresses are absolute, not relative.
 *
 *  @return Data read. Only the lower 16 bits are meaningful.
 */
unsigned int cflash_read(unsigned int address) {
	unsigned char *p;
	unsigned short s;
	int i;
	UINT cluster,cluster2,cluster3,fileLBA;

	//wsprintf(buffer,"Reading from %08X\n",address);
	//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);

	// Status register
	if (address == CF_REG_STS) {
		s = cf_reg_sts;
		return s; 

	// Data register
	} else if (address == CF_REG_DATA) {
		cluster = (currLBA / (512 * SECPERCLUS));
		cluster2 = (((currLBA/512) - filesysData) / SECPERCLUS) + 2;

		if (cf_reg_cmd == CMD_READ) {
			// Reading from the MBR 
			if (currLBA < 512) {
				p = (unsigned char*)&MBR;
				s = *(unsigned short *)(p + currLBA);

			// Reading the FAT 
			} else if ((currLBA >= filesysFAT*512) && (currLBA < filesysRootDir*512)) {
				p = (unsigned char*)&FAT16[0];
				s = *(unsigned short *)(p + (currLBA-filesysFAT*512));

			// Reading directory entries 
			} else if ((currLBA >= filesysRootDir*512) &&
				       (cluster <= lastDirEntCluster)) {
				cluster3 = ((currLBA - (SECRESV * 512)) / (512 * SECPERCLUS));
				i = (currLBA-(((cluster3-(filesysRootDir/SECPERCLUS))*SECPERCLUS+filesysRootDir)*512)); //(currLBA - cluster*BYTESPERCLUS);
				//if ((currLBA-(((cluster-(filesysRootDir/SECPERCLUS))*SECPERCLUS+filesysRootDir)*512)) < (dirEntriesInCluster[cluster]*32)) {
				if (i < (dirEntriesInCluster[cluster3]*32)) {
					//i = (currLBA-(((cluster-(filesysRootDir/SECPERCLUS))*SECPERCLUS+filesysRootDir)*512));
					//wsprintf(buffer,"Reading from dirent %d in cluster %d (%08X)\n",i/32,cluster3,currLBA);
					//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
					
					p = (unsigned char*)dirEntryPtr[cluster3];
					s = *(unsigned short*)(p + i);
				} else {
					//i = (currLBA-(((cluster-(filesysRootDir/SECPERCLUS))*SECPERCLUS+filesysRootDir)*512)) - (dirEntriesInCluster[cluster]*32);
					i /= 32;
					i -= dirEntriesInCluster[cluster3];
					if ((i>=0)&&(i<numExtraEntries[cluster3])) {
						//wsprintf(buffer,"Reading from extra dirent %d in cluster %d (%08X)\n",i,cluster3,currLBA);
						//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
						p = (unsigned char*)extraDirEntries[cluster3];
						s = *(unsigned short*)(p + i*32 + (currLBA&0x1F));						
						//wsprintf(buffer,"Reading %04X from extra entry #%d in cluster %d (%08X)\n",s,i,cluster,currLBA);
						//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
					} else if ((currLBA&0x1F)==0) {
						s = FILE_FREE;
					} else {
						s = 0;
					}
				}
			// Reading file data 
			} else if ((cluster2 > lastDirEntCluster) && (cluster2 <= lastFileDataCluster)) { //else if ((cluster>lastDirEntCluster)&&(cluster<=lastFileDataCluster)) {
				fileLBA = currLBA - (filesysData-32)*512;	// 32 = # sectors used for the root entries

				//if (activeDirEnt==27) {
				//wsprintf(buffer,"%d, %d [%d, %d, %d]\n",fileLBA,currLBA,firstDirEntCluster,lastDirEntCluster,lastFileDataCluster);
				//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
				//}

				// Check if the read is from the currently opened file 
				if ((fileLBA >= fileStartLBA) && (fileLBA < fileEndLBA)) {
					cluster = (fileLBA / (512 * SECPERCLUS));
					s = fread_buffered(activeDirEnt,cluster-dirEntries[activeDirEnt].startCluster,(fileLBA-fileStartLBA)&(BYTESPERCLUS-1)); 
				} else {
					for (i=0; i<numFiles; i++) {
						if ((fileLBA>=(dirEntries[i].startCluster*512*SECPERCLUS)) &&
							(fileLBA <(dirEntries[i].startCluster*512*SECPERCLUS)+dirEntries[i].fileSize) &&
							((dirEntries[i].attrib & (ATTRIB_DIR|ATTRIB_LFN))==0)) {
							cluster = (fileLBA / (512 * SECPERCLUS));
							s = fread_buffered(i,cluster-dirEntries[i].startCluster,fileLBA&(BYTESPERCLUS-1));
							break;
						}
					}
				}
			} 
			currLBA += 2;
			return s;
		}

	// Command register
	} else if (address == CF_REG_CMD) {
		s = 0;
		return 0;


	} else if (address == CF_REG_LBA1) {
		s = cf_reg_lba1;
		return s; 
	}

	return 0;
}





/**
 *  @brief Writes to one of the CF device registers
 *
 *  @param address Address to write to. Addresses are absolute, not relative.
 *  @param data    Data to write to the given address. Only the lower 16 bits are used.
 *
 *  @return
 */
void cflash_write(unsigned int address,unsigned int data) {
	UINT cluster,cluster2,cluster3,fileLBA;
	int i,j,k,l;
	unsigned char *p;
	unsigned short s;
	DIR_ENT *d;

	//wsprintf(buffer,"Writing %X to %08X\n",data,address);
	//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);

	// Status register
	if (address == CF_REG_STS) {
		cf_reg_sts = data & 0xFFFF;

	// Data register
	} else if (address == CF_REG_DATA) {
		cluster = (currLBA / (BYTESPERSEC * SECPERCLUS));
		cluster2 = (((currLBA / BYTESPERSEC) - filesysData) / SECPERCLUS) + 2;
		s = (unsigned short)data;

		if (cf_reg_cmd == CMD_WRITE) {
			//return;

			// Writing to the MBR (not allowed) 
			if (currLBA < 512) {
				//p = (unsigned char*)&MBR;
				//s = *(unsigned short *)(p + currLBA);

			// Writing to the FAT 
			} else if ((currLBA >= filesysFAT * 512) && (currLBA < filesysRootDir * 512)) {
				p = (unsigned char *)&FAT16[0];
				*(unsigned short *)(p + (currLBA-filesysFAT*512)) = s;
				/*wsprintf(buffer,"Writing %X to FAT (%08X)\n",s,currLBA);
				WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);*/

			// Writing directory entries 
			} else if ((currLBA >= filesysRootDir*512) &&
				       (cluster <= lastDirEntCluster)) {
				/*if ((currLBA-(((cluster-(filesysRootDir/SECPERCLUS))*SECPERCLUS+filesysRootDir)*512)) < (dirEntriesInCluster[cluster]*32)) {
					p = (unsigned char*)dirEntryPtr[cluster];
					*(unsigned short*)(p + (currLBA-(((cluster-(filesysRootDir/SECPERCLUS))*SECPERCLUS+filesysRootDir)*512))) = s;

					if ((currLBA>=0x10700)&&(currLBA<0x10720)) {
						wsprintf(buffer,"Writing %X to DIRENT (%08X), cluster %d\n",s,currLBA,cluster);
						WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
					}*/
				cluster3 = ((currLBA - (SECRESV * 512)) / (512 * SECPERCLUS));
				i = (currLBA-(((cluster3-(filesysRootDir/SECPERCLUS))*SECPERCLUS+filesysRootDir)*512)); //(currLBA - cluster*BYTESPERCLUS);
				if (i < (dirEntriesInCluster[cluster3]*32)) {
					p = (unsigned char*)dirEntryPtr[cluster3];
					*(unsigned short*)(p + i) = s;
					//wsprintf(buffer,"+Writing %X to DIRENT (%08X), cluster %d\n",s,currLBA,cluster3);
					//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
				} else {
					//i = (currLBA-(((cluster-(filesysRootDir/SECPERCLUS))*SECPERCLUS+filesysRootDir)*512)) - (dirEntriesInCluster[cluster]*32);
					i /= 32;
					i -= dirEntriesInCluster[cluster3];
					if (i<64) { //((dirEntriesInCluster[cluster]+i)<=256) {
						//wsprintf(buffer,"Adding %d entries to cluster\n",i+1);
						//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
						//wsprintf(buffer,"Writing %X to DIRENT (%08X), cluster %d\n",s,currLBA,cluster3);
						//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
						
						p = (unsigned char*)extraDirEntries[cluster3];
						if (p==NULL)
							extraDirEntries[cluster3] = (DIR_ENT*)LocalAlloc(LPTR,64*32);
						p = (unsigned char*)extraDirEntries[cluster3];
						*(unsigned short*)(p + i*32 + (currLBA&0x1F)) = s;
						//dirEntriesInCluster[cluster] += i;
						//numFiles += i;
						if ((i+1) > numExtraEntries[cluster3])
							numExtraEntries[cluster3] = i+1;
					}
				}

			// Writing file data 
			} else if (cluster2 > lastDirEntCluster) { //&& (cluster2 <= lastFileDataCluster)) { //else if ((cluster>lastDirEntCluster)&&(cluster<=lastFileDataCluster)) {
				fileLBA = currLBA - (filesysData-32)*512;	// 32 = # sectors used for the root entries

				//wsprintf(buffer,"Writing %X to file (%08X, %08X)\n",s,currLBA,fileLBA);
				//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);

				cluster = (fileLBA / (512 * SECPERCLUS));
				j = cluster;
				k = 0;
				do {
					for (i=0; i<MAXFILES; i++) {
						if (FAT16[i] == j) {
							j = i;
							k++;
							break;
						}
					}
				} while (i < MAXFILES);
				for (i=0; i<numFiles; i++) {
					if (dirEntries[i].startCluster == j) {
						if ((dirEntries[i].attrib & (ATTRIB_LFN|ATTRIB_DIR))==0) {
							l = -2;
							break;
						}
					}
				}
				if (i>=numFiles) {
					for (l=0; l<MAXFILES; l++) {
						d = extraDirEntries[l];
						for (i=0; i<numExtraEntries[l]; i++) {
							if (d[i].startCluster == j) {
								if ((d[i].attrib & (ATTRIB_LFN|ATTRIB_DIR))==0) {
									l = -1;
									break;
								}
							}
						}
						if (l == -1)
							break;
					}
				}
				//wsprintf(buffer,"Cluster %d is #%d in the chain. Starting cluster is %d. LBA=%08X, extra=%d\n",cluster,k,j,fileLBA,numExtraEntries[cluster]);
				//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);

				//for (i=0; i<numExtraEntries[cluster]; i++) {
				//	d = extraDirEntries[cluster];
				//	if ((j == d[i].startCluster) &&
				//		((d[i].attrib & (ATTRIB_LFN|ATTRIB_DIR))==0)) {
						//cluster = (fileLBA / (512 * SECPERCLUS));
				if (l == -1) {
					//wsprintf(buffer,"|Writing %X to file (%08X, %08X)\n",s,currLBA,fileLBA);
					//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
					fwrite_buffered(i, &d[i], cluster-d[i].startCluster, fileLBA&(BYTESPERCLUS-1), k, data);
					//break;
				} else if (l == -2) {
					fwrite_buffered(i, NULL, cluster-dirEntries[i].startCluster, fileLBA&(BYTESPERCLUS-1), k, data);
				}
			

				// Check if the read is from the currently opened file 
				/*if ((fileLBA >= fileStartLBA) && (fileLBA < fileEndLBA)) {
					cluster = (fileLBA / (512 * SECPERCLUS));
					fwrite_buffered(activeDirEnt,cluster-dirEntries[activeDirEnt].startCluster,(fileLBA-fileStartLBA)&(BYTESPERCLUS-1),data); 
				} else {*/
				/*
					for (i=0; i<numFiles; i++) {
						if ((fileLBA>=(dirEntries[i].startCluster*512*SECPERCLUS)) &&
							//(fileLBA <(dirEntries[i].startCluster*512*SECPERCLUS)+dirEntries[i].fileSize) &&
							((dirEntries[i].attrib & (ATTRIB_LFN|ATTRIB_DIR))==0)) {
							if ((i<numFiles-1)&&(fileLBA >=(dirEntries[i+1].startCluster*512*SECPERCLUS)+dirEntries[i].fileSize))
								continue;
							cluster = (fileLBA / (512 * SECPERCLUS));
							wsprintf(buffer,"Writing %X to file (%08X, %08X)\n",s,currLBA,fileLBA);
							//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
							fwrite_buffered(i,cluster-dirEntries[i].startCluster,fileLBA&(BYTESPERCLUS-1),data);
							break;
						}
					}
					for (i=0; i<numExtraEntries[cluster]; i++) {
						d = extraDirEntries[cluster];
						if ((fileLBA>=(d[i].startCluster*512*SECPERCLUS)) &&
							//(fileLBA <(d[i].startCluster*512*SECPERCLUS)+d[i].fileSize) &&
							((d[i].attrib & (ATTRIB_LFN|ATTRIB_DIR))==0)) {
							cluster = (fileLBA / (512 * SECPERCLUS));
							wsprintf(buffer,"|Writing %X to file (%08X, %08X)\n",s,currLBA,fileLBA);
							//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
							fwrite_buffered(i,cluster-d[i].startCluster,fileLBA&(BYTESPERCLUS-1),data);
							break;
						}
					}
					*/

				//}
			} 
			currLBA += 2;
		}

	// COmmand register
	} else if (address==CF_REG_CMD) {
		cf_reg_cmd = data&0xFF;
		cf_reg_sts = 0x58;	// READY
	
	} else if (address==CF_REG_LBA1) {
		cf_reg_lba1 = data&0xFF;
		currLBA = (currLBA&0xFFFFFF00)|cf_reg_lba1;
	
	} else if (address==CF_REG_LBA2) {
		cf_reg_lba2 = data&0xFF;
		currLBA = (currLBA&0xFFFF00FF)|(cf_reg_lba2<<8);
	
	} else if (address==CF_REG_LBA3) {
		cf_reg_lba3 = data&0xFF;
		currLBA = (currLBA&0xFF00FFFF)|(cf_reg_lba3<<16);
	
	} else if (address==CF_REG_LBA4) {
		cf_reg_lba4 = data&0xFF;
		currLBA = (currLBA&0x00FFFFFF)|((cf_reg_lba4&0x0F)<<24);
		currLBA *= 512;
	}
}



/**
 *  @brief Shuts down CFlash emulation
 *
 *  @return
 */
void cflash_close() {
	int i;

	if (cflashDeviceEnabled) {
		cflashDeviceEnabled = false;

		for (i=0; i<MAXFILES; i++) {
			if (extraDirEntries[i] != NULL)
				LocalFree(extraDirEntries[i]);
		}

		if (dirEntries != NULL)
			LocalFree((HLOCAL)dirEntries);

		if (dirEntryLink != NULL)
			LocalFree((HLOCAL)dirEntryLink);

		if (activeDirEnt != -1)
			CloseHandle(hFile);

		if (gpfWrite != NULL) {
			fclose(gpfWrite);
			gpfWrite = NULL;
		}
	}
}
